import torch
from tqdm import tqdm
from transformers import AutoTokenizer, AutoModelForSequenceClassification, AdamW, get_scheduler, TrainingArguments, Trainer
import argparse
import pickle as pkl
import random
from sklearn import metrics
import os
import numpy as np
#import accelerate
import evaluate
from torch.utils.data import DataLoader, RandomSampler, SequentialSampler, Dataset

device=torch.device("cuda" if torch.cuda.is_available() else "cpu")
#from accelerate import Accelerator

#accelerator=Accelerator()
parser=argparse.ArgumentParser()

class AKIDataset(torch.utils.data.Dataset):
    def __init__(self, instances, tokenizer):
        self.instances=instances
        self.tokenizer=tokenizer
    def __getitem__(self, idx):
        instance=self.instances[idx]
        #print(instance)
        prompt=instance[0]
        labels=instance[1]
        item=self.tokenize(prompt)
        item["input_ids"]=item["input_ids"].squeeze()
        item["token_type_ids"]=item["token_type_ids"].squeeze()
        item["attention_mask"]=item["attention_mask"].squeeze()
        item["label"] = labels
        #print(item["input_ids"].size())
        #print(item["input_ids"].size())
        #item["input_ids"]
        #item["label"]#.to(device)
        #item["attention_mask"]
        return item

    def tokenize(self, prompt):
        result=self.tokenizer(prompt, truncation=True, max_length=512, padding="max_length", return_tensors="pt")
        return result
    def __len__(self):
        return len(self.instances)

def set_seed(seed):
    np.random.seed(seed)
    random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.backends.cudnn.deterministic=True
    torch.backends.cudnn.banchmark=False
    os.environ["PYTHONHASHSEED"]=str(seed)
    print("Random seed set as : {seed}")


def train(model, dataset, seed, batch_size, epochs, lr, gradient_accumulation_steps=1, model_save_path=None):
    model.train()
    model.to(device)
    set_seed(seed)
    optimizer=AdamW(model.parameters(), lr=lr)
    train_sampler=RandomSampler(dataset)
    train_dataloader=DataLoader(dataset, sampler=train_sampler, batch_size=batch_size)
    num_training_steps=epochs*len(train_dataloader)
    lr_scheduler=get_scheduler(name="linear", optimizer=optimizer, num_warmup_steps=0, num_training_steps=num_training_steps)

    #model, optimizer, train_dataloader, lr_scheduler=accelerator.prepare(model, optimizer, train_dataloader, lr_scheduler)
    model.zero_grad()

    for e in tqdm(range(epochs)):
        model.train()
        for index, batch in tqdm(enumerate(train_dataloader)):
            #print(batch)
            input_ids=batch["input_ids"].squeeze().to(device)#.to(device) #.unsqueeze(0).transpose(0,1).to(device)
            attention_mask=batch["attention_mask"].squeeze().to(device)#.to(device) #).unsqueeze(0).transpose(0,1).to(device)
            token_type_ids=batch["token_type_ids"].squeeze().to(device)
            labels=batch["label"].to(device) #).unsqueeze(0).transpose(0,1).to(device)
            print(input_ids.size())
            outputs=model(input_ids=input_ids, attention_mask=attention_mask, token_type_ids=token_type_ids, labels=labels)
            print(outputs.logits)
            print("Model prediciton: ", torch.argmax(outputs.logits, dim=-1))
            print("True label: ", labels)

            loss=outputs.loss
            loss=loss/gradient_accumulation_steps
            loss.backward()
            print(loss)
            #if ((index+1) % gradient_accumulation_steps==0) or (index+1==len(train_dataloader)):
             #   optimizer.step()
              #  lr_scheduler.step()
               # optimizer.zero_grad()
            if index % gradient_accumulation_steps==0 or index == len(train_dataloader):
                optimizer.step()
                lr_scheduler.step()
                optimizer.zero_grad()
            #model.zero_grad()
    #if model_save_path != None:
     #   torch.save(model, model_save_path)
    model.save_pretrained(model_save_path)
	
    return model


def main():
    metric=evaluate.load("accuracy", average="macro")
	#metric_f1=evaluate.load("f1")
    model=AutoModelForSequenceClassification.from_pretrained(args.model, num_labels=3)
    tokenizer=AutoTokenizer.from_pretrained(args.tokenizer)
    with open(args.train_dataset, "rb") as f:
        data=pkl.load(f)
    print(len(data[:-1]))
    with open(args.test_dataset, "rb") as f:
        test_data=pkl.load(f)
    test_dataset=AKIDataset(test_data, tokenizer)
    dataset=AKIDataset(data[:-1], tokenizer)
    training_args=TrainingArguments(output_dir="./outputs_3_eps_full",
                    per_device_train_batch_size=args.train_batch_size, 
                    gradient_accumulation_steps=2,
                    learning_rate=args.learning_rate,
					evaluation_strategy="epoch")
    trainer=Trainer(model=model,
                     args=training_args,
                     tokenizer=tokenizer,
                     train_dataset=dataset, 
                     eval_dataset=test_dataset,
                     compute_metrics=compute_metrics)
    trainer.train()

    #trained_model=train(model, dataset, args.seed, args.train_batch_size, args.epochs, args.learning_rate, args.gradient_accumulation_steps, args.model_save_path)

def compute_metrics(eval_pred):
	#metric=evaluate.load("accuracy", average="macro")
	#metric_f1=evaluate.load("f1", average="macro")
	
	#metric_prec=evaluate.load("precision", average="macro")
	#metric_rec=evaluate.load("recall", average="macro")
	#print(eval_pred)
	logits, labels=eval_pred
	#print(logits, labels)
	normalized_logits=torch.nn.functional.log_softmax(torch.from_numpy(logits), dim=1)
	predictions=torch.argmax(normalized_logits, dim=-1)
	
	#acc=metric.compute(predictions=predictions, references=labels)
	#f1=metric_f1.compute(predictions=predictions, references=labels)
	#precision=metric_prec.compute(predictions=predictions, references=labels)
	#recall=metric_rec.compute(predictions=predictions, references=labels)
	
	accuracy=metrics.accuracy_score(labels, predictions)
	f1_score_micro=metrics.f1_score(labels, predictions, average="micro")
	f1_score_macro=metrics.f1_score(labels, predictions, average="macro")
	print(f"Accuracy score: {accuracy}")
	print(f"F1 score (micro): {f1_score_micro}")
	print(f"F1 score (macro): {f1_score_macro}")
	return_dict={"accuracy":accuracy, "f1_score_micro":f1_score_micro, "f1_score_macro":f1_score_macro}
	return return_dict#, precision, recall

if __name__ == "__main__":
    parser.add_argument("--model", type=str)
    parser.add_argument("--tokenizer", type=str)
    parser.add_argument("--train_dataset", type=str)
    parser.add_argument("--test_dataset", type=str)
    parser.add_argument("--seed", type=int)
    parser.add_argument("--epochs", type=int)
    parser.add_argument("--learning_rate", type=float)
    parser.add_argument("--train_batch_size", type=int)
    parser.add_argument("--gradient_accumulation_steps", type=int)
    parser.add_argument("--model_save_path", type=str)

    args=parser.parse_args()
    main()


